// Michael J. Wilson
// Music 256a Fall 2010
// Homework #3

#ifndef __AUDIO_TRANSFORM_H__
#define __AUDIO_TRANSFORM_H__

#include <vector>
#include <string>
#include <pthread.h>

class WaterfallBuffer;

enum AUDIO_WINDOW_TYPE
{
    RECTANGULAR,
    HANNING,
    HAMMING,
    BLACKMAN
};

/// Contains methods for storing and transforming audio data
class AudioTransform
{
public:
    AudioTransform(unsigned int buffer_size, unsigned int num_waterfalls);
    ~AudioTransform();

    /// Accept new samples into the buffer.  Automatically computes
    /// frames, etc.
    void accept_samples(float* samples, unsigned int num_samples);

    /// Set the window type to use
    void set_window_type(AUDIO_WINDOW_TYPE window_type);

    /// Flag set to true when there is new data to display
    bool new_data_is_available();

    /// Get pointer to windowed data, should call
    /// start_consuming_data() first.
    std::vector<float>* get_windowed_data();

    /// Get pointer to specified magnitude spectrum, should call
    /// start_consuming_data() first.
    std::vector<float>* get_magnitude_spectrum(unsigned int index = 0);

    /// Get a pitch guess, based on autocorrelation of the last
    /// time-domain datapoints, should call start_consuming_data()
    /// first.
    std::string get_pitch_guess();

    /// Locks data and resets new data flag
    void start_consuming_data();

    /// Unlocks data
    void stop_consuming_data();

    /// Number of waterfalls
    unsigned int num_waterfalls();

private:
    /// Mutex for graphics operations.  Should be locked when
    /// specified in a function definition.
    pthread_mutex_t graphics_mutex;

    /// Graphics buffer.  You should lock the mutex before reading from or writing to this.
    std::vector<float> graphics_buffer;

    /// Internal data buffers for computation
    std::vector<float> fft_buffer;
    std::vector<float> windowed_data;

    /// Magnitude frequency spectrums
    WaterfallBuffer* m_magnitude_spectra;

    /// Hamming window
    std::vector<float> m_hamming_window;
    /// Hanning window
    std::vector<float> m_hanning_window;
    /// Blackman window
    std::vector<float> m_blackman_window;
    /// Currently selected window type
    AUDIO_WINDOW_TYPE m_window_type;

    /// Current sample in the audio input buffer
    unsigned int m_current_sample;

    /// Buffer sizes
    unsigned int m_buffer_size;
    unsigned int m_fft_size;

    /// Flag for new data available
    bool m_new_data_is_available;

    /// The last pitch that was guessed
    std::string m_pitch_guess;

    /// Transforms the audio buffer and notes that it's ready to be displayed
    void transform_buffer();
};

#endif
