// Michael J. Wilson
// Music 256a Fall 2010
// Homework #3

#include "RtAudio.h"
#include "Utility.h"
#include "AudioTransform.h"
#include <iostream>
#include <cmath>
#include <cstdlib>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <pthread.h>

/// For debugging
const bool DEBUG = false;

/// Global audio transformation object
AudioTransform* g_audio = NULL;

/// Audio parameters
bool zoom = false;
bool db_scale = false;
bool adaptive_linear_scale = false;
bool display_pitch_guess = true;
float max_magnitude = Utility::EPSILON;

//////////////////////////////////////////////////////////////////////
/// Audio callback
//////////////////////////////////////////////////////////////////////
int audio_callback(void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames, double streamTime, RtAudioStreamStatus status, void *userData)
{
    float* ibuffer = (float*)inputBuffer;

    if (status)
    {
        std::cout << "Stream overflow detected!" << std::endl;
    }

    // Fill graphics buffer
    g_audio->accept_samples(ibuffer, nBufferFrames);
    
    glutPostRedisplay();

    return(0);
}

//////////////////////////////////////////////////////////////////////
/// Idle callback
//////////////////////////////////////////////////////////////////////
void idle_func()
{
}

//////////////////////////////////////////////////////////////////////
/// Display callback
//////////////////////////////////////////////////////////////////////
void display_func()
{
    // Options we can allow to change at runtime:
    // - Window type
    // - Scale on spectrum (db, magnitude, adaptive, etc)
    // - Colors?  Something else?

    // Need also to visualize something else

    // Lower alignment bound for dB spectral magnitude display
    const float db_truncate = -100;
    const float db_lower_bound = -300;

    // Drawing offsets
    const GLfloat raw_offset = 2.5;
    const GLfloat spectral_offset = -2.5;

    // Scale factor for waveforms
    const float raw_scale = 5.0f;
    const float spectral_scale = 5.0f;
    const float zoom_factor = 3;

    // Text for pitch guess
    const std::string pitch("Pitch guess: ");

    unsigned int i;
    unsigned int waterfall_index;
    GLfloat x;
    GLfloat y;
    GLfloat z;
    GLfloat xinc;
    GLfloat zinc;
    std::string note_value;

    if (g_audio->new_data_is_available())
    {
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        glLineWidth( 1.0 );
        glPushMatrix();

        // Notify the audio transform that we're consuming data
        g_audio->start_consuming_data();

        // Time data
        x = -5;
        xinc = fabs(2*x / g_audio->get_windowed_data()->size());
        glColor3f(0.0, 1.0, 1.0);
        glBegin(GL_LINE_STRIP);
        for(i = 0; i < g_audio->get_windowed_data()->size(); i++)
        {
            // Update max magnitude value since we're already looping here:
            if (adaptive_linear_scale)
            {
                if ((*g_audio->get_magnitude_spectrum())[i] > max_magnitude)
                {
                    max_magnitude = (*g_audio->get_magnitude_spectrum())[i];
                }
            }
            // Draw next vertex
            y = (raw_scale / 2) * (*g_audio->get_windowed_data())[i];
            glVertex2f(x, raw_offset + y);
            x += xinc;
        }
        glEnd();

        // Frequency data
        xinc = fabs(2*x / g_audio->get_windowed_data()->size());
        if (zoom)
        {
            xinc *= zoom_factor;
        }
        z = 0;
        zinc = -0.5;
        for (waterfall_index = 0; waterfall_index < g_audio->num_waterfalls(); waterfall_index++)
        {
            x = -5;
            glBegin(GL_LINE_STRIP);
            for(i = 0; i < (*g_audio->get_magnitude_spectrum(waterfall_index)).size(); i++)
            {
                if (db_scale)
                {
                    // Trucnated dB scale
                    y = (20.0f * std::log10((*g_audio->get_magnitude_spectrum(waterfall_index))[i]));
                    if (y < db_lower_bound)
                    {
                        y = db_lower_bound;
                    }
                    y -= db_truncate;
                    y /= -db_truncate;
                    y *= spectral_scale;
                }
                else
                {
                    // Linear scale
                    y = (*g_audio->get_magnitude_spectrum(waterfall_index))[i] * spectral_scale;
                    if (adaptive_linear_scale)
                    {
                        y /= max_magnitude;
                    }
                    else
                    {
                        y *= 10;
                    }
                }
                glColor3f(1.0 / (float)(waterfall_index + 1), 1.0 / (float)(waterfall_index + 1), 0.0);
                glVertex3f(x, spectral_offset + y, z);
                x += xinc;
            }
            glEnd();
            z += zinc;
        }

        // Print pitch guess
        if (display_pitch_guess)
        {
            glColor3f(1.0, 0.0, 1.0);
            glTranslatef(-1, 0, 0);
            glScalef(.0025, .0025, .0025);
            for( i = 0; i < pitch.length(); i++ )
            {
                glutStrokeCharacter(GLUT_STROKE_ROMAN, pitch[i]);
            }
            note_value = g_audio->get_pitch_guess();
            for( i = 0; i < note_value.length(); i++ )
            {
                glutStrokeCharacter(GLUT_STROKE_ROMAN, note_value[i]);
            }
        }

        // Notify the audio transform that we're done consuming data
        g_audio->stop_consuming_data();

        glPopMatrix();
        glFlush();
        glutSwapBuffers();
    }
}

//////////////////////////////////////////////////////////////////////
/// Window reshape callback
//////////////////////////////////////////////////////////////////////
void reshape_func(int width, int height)
{
    glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45.0, (GLfloat) width / (GLfloat) height, 1.0, 300.0 );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}

//////////////////////////////////////////////////////////////////////
/// Keyboard callback
//////////////////////////////////////////////////////////////////////
void keyboard_func(unsigned char key, int x, int y)
{
    switch (key)
    {
    case 'Q':
    case 'q':
        exit(EXIT_SUCCESS);
        break;
    case 'D':
    case 'd':
        db_scale = !db_scale;
        break;
    case 'B':
    case 'b':
        g_audio->set_window_type(BLACKMAN);
        break;
    case 'N':
    case 'n':
        g_audio->set_window_type(HANNING);
        break;
    case 'M':
    case 'm':
        g_audio->set_window_type(HAMMING);
        break;
    case 'R':
    case 'r':
        g_audio->set_window_type(RECTANGULAR);
        break;
    case 'A':
    case 'a':
        adaptive_linear_scale = !adaptive_linear_scale;
        max_magnitude = Utility::EPSILON;
        break;
    case 'Z':
    case 'z':
        zoom = !zoom;
        break;
    case 'P':
    case 'p':
        display_pitch_guess = !display_pitch_guess;
        break;
    }
}

//////////////////////////////////////////////////////////////////////
/// Special key callback
//////////////////////////////////////////////////////////////////////
void special_func(int key, int x, int y)
{
}

//////////////////////////////////////////////////////////////////////
/// Exit function
//////////////////////////////////////////////////////////////////////
void exit_function()
{
    //////////////////////////////////////////////////
    // Cleanup
    if (Utility::get_utility()->dac->isStreamRunning())
    {
        try
        {
            // Stop the stream
            Utility::get_utility()->dac->stopStream();
        }
        catch (RtError& e)
        {
            e.printMessage();
        }
    }

    if (Utility::get_utility()->dac->isStreamOpen())
    {
        Utility::get_utility()->dac->closeStream();
    }

    delete g_audio;
}

//////////////////////////////////////////////////////////////////////
/// Handle main input
//////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    bool error_occured = false;

    unsigned int waveform_buffer_size = 1024;
    unsigned int num_waterfalls = 20;
    unsigned int buffer_frames = 1024;

    //////////////////////////////////////////////////
    // Set up exit function, since GLUT doesn't have
    // provisions for an exit function
    std::atexit(exit_function);

    //////////////////////////////////////////////////
    // Set up graphics
    //
    // NOTE - glutInit() should be called before we parse our own
    // options out, since it extracts options.
    glutInit(&argc, argv);
    glutInitWindowSize(Utility::WINDOW_WIDTH, Utility::WINDOW_HEIGHT);
    glutInitWindowPosition(-1, -1);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );

    //////////////////////////////////////////////////
    // Create graphics window
    glutCreateWindow(Utility::PROGRAM_TITLE.c_str());

    //////////////////////////////////////////////////
    // Set up GLUT callbacks
    glutDisplayFunc(display_func);
    glutReshapeFunc(reshape_func);
    glutKeyboardFunc(keyboard_func);
    glutMouseFunc(NULL);
    glutMotionFunc(NULL);
    glutPassiveMotionFunc(NULL);
    glutVisibilityFunc(NULL);
    glutEntryFunc(NULL);
    glutSpecialFunc(special_func);
    glutMenuStatusFunc(NULL);
    glutIdleFunc(idle_func);

    //////////////////////////////////////////////////
    // Enable OpenGL options
    glEnable( GL_DEPTH_TEST );
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);

    //////////////////////////////////////////////////
    // Add a light roughly around the eye area
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable( GL_COLOR_MATERIAL );
    float position[] = {0.0, 0.0, 1.0, 1.0};
    float color[] = {1.0, 1.0, 1.0, 0.5};
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glLightfv(GL_LIGHT0, GL_AMBIENT, color);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
    glLightfv(GL_LIGHT0, GL_SPECULAR, color);
    glEnable(GL_LIGHT0);

    //////////////////////////////////////////////////
    // Set up audio
    if (Utility::get_utility()->dac->getDeviceCount() < 1)
    {
        std::cout << std::endl << "No audio devices found!" << std::endl;
        error_occured = true;
    }
    if (!error_occured)
    {
        // Get the best sample rate, up to our maximum:
        Utility::sample_rate = 0;
        RtAudio::DeviceInfo info = Utility::get_utility()->dac->getDeviceInfo(Utility::get_utility()->dac->getDefaultInputDevice());
        for (std::vector<unsigned int>::iterator sr_iter = info.sampleRates.begin(); sr_iter != info.sampleRates.end(); ++sr_iter)
        {
            if (*sr_iter <= Utility::MAXIMUM_SAMPLE_RATE && *sr_iter > Utility::sample_rate)
            {
                Utility::sample_rate = *sr_iter;
            }
        }

        RtAudio::StreamParameters input_parameters;
        input_parameters.deviceId = Utility::get_utility()->dac->getDefaultInputDevice();
        input_parameters.nChannels = 1;
        input_parameters.firstChannel = 0;

        try
        {
            Utility::get_utility()->dac->openStream(NULL,
                           &input_parameters,
                           RTAUDIO_FLOAT32,
                           Utility::sample_rate,
                           &buffer_frames,
                           &audio_callback,
                           NULL);
            Utility::get_utility()->dac->startStream();
            std::cout << "Opened audio interface" << std::endl
                      << "Sample rate: " << Utility::sample_rate << std::endl
                      << "Frames: " << buffer_frames << std::endl;


            //////////////////////////////////////////////////
            // Create audio object
            g_audio = new AudioTransform(waveform_buffer_size, num_waterfalls);
        }
        catch (RtError& e)
        {
            e.printMessage();
            error_occured = true;
        }
    }

    //////////////////////////////////////////////////
    // Start main loop
    if (!error_occured)
    {
        // Print command list to terminal
        std::cout << std::endl
                  << "Commands:" << std::endl
                  << "q - Quit" << std::endl
                  << "d - Toggle dB or linear spectral magnitude scale" << std::endl
                  << "b - Set window type to Blackman" << std::endl
                  << "n - Set window type to Hanning" << std::endl
                  << "m - Set window type to Hamming" << std::endl
                  << "r - Set window type to Rectangular" << std::endl
                  << "a - Toggle adaptive leveling on linear spectral magnitude scale" << std::endl
                  << "z - Zoom spectral magnitude scale" << std::endl
                  << "p - Toggle pitch guess display" << std::endl;

        // Run main loop
        glutMainLoop();
    }

    return (error_occured ? EXIT_FAILURE : EXIT_SUCCESS);
}
